// document class

#include "stdafx.h"
#include "doc.h"
#include "midi.h"
#include "midifish.h"



// construct
Doc::Doc()
{
	int i;

	for (i=0; i<NUM_INS; i++)
		mIn[i] = new MidiInput(this);

	for (i=0; i<NUM_OUTS; i++)
		mOut[i] = new MidiOutput(this);

	for (i=0; i<NUM_KEYS; i++)
		mKeyMap[i] = new KeyMap(i);

	for (i=0; i<NUM_CTRLS; i++) {
		mCtrlMap[i] = new CtrlMap(i);

		mCtrlIn[i] = 0;
		mCtrlOut[i] = 0;
	}

	mLastNote = 0;

	mThruPort = 0;

	mKeyMapCtrl = NULL;
	mKeyActiveCtrl = NULL;

	mCtrlMapCtrl = NULL;
	mCtrlActiveCtrl = NULL;

	reset();

	mFilename = "";

	mDirty = false;
}



// destruct
Doc::~Doc()
{
	int i;

	for (i=0; i<NUM_INS; i++)
		delete mIn[i];

	for (i=0; i<NUM_OUTS; i++)
		delete mOut[i];

	for (i=0; i<NUM_KEYS; i++)
		delete mKeyMap[i];

	for (i=0; i<NUM_CTRLS; i++)
		delete mCtrlMap[i];
}



// reset data
void Doc::reset()
{
	reset_keys();
	reset_controllers();
}



// reset the key mapping
void Doc::reset_keys()
{
	for (int i=0; i<NUM_KEYS; i++)
		*mKeyMap[i] = KeyMap(i);

	mKeyActive = true;

	mDirty = true;

	if (mKeyMapCtrl)
		update_key_controls();
}



// reset the controller mapping
void Doc::reset_controllers()
{
	for (int i=0; i<NUM_CTRLS; i++)
		*mCtrlMap[i] = CtrlMap(i);

	mCtrlActive = true;

	mDirty = true;

	if (mCtrlMapCtrl)
		update_ctrl_controls();
}



// read a file
void Doc::load(const string &filename)
{
	int i;

	FILE *file = fopen(filename.c_str(), "rb");

	if ((!file) || (fgetl(file) != 0xE1F23E1F)) {
		AfxMessageBox(CString("Error reading\n") + filename.c_str());
		return;
	}

	for (i=0; i<NUM_INS; i++)
		mIn[i]->load(file);

	for (i=0; i<NUM_OUTS; i++)
		mOut[i]->load(file);

	mKeyActive = fgetc(file);
	mCtrlActive = fgetc(file);

	for (i=0; i<NUM_CTRLS; i++)
		mCtrlMap[i]->load(file);

	for (i=0; i<NUM_KEYS; i++)
		mKeyMap[i]->load(file);

	fclose(file);

	mFilename = filename;

	mDirty = false;

	if (mKeyMapCtrl)
		update_key_controls();

	if (mCtrlMapCtrl)
		update_ctrl_controls();
}



// write a file
void Doc::save(const string &filename)
{
	int i;

	FILE *file = fopen(filename.c_str(), "wb");

	if (!file) {
		AfxMessageBox(CString("Error writing\n") + filename.c_str());
		mDirty = true;
		return;
	}

	fputl(0xE1F23E1F, file);

	for (i=0; i<NUM_INS; i++)
		mIn[i]->save(file);

	for (i=0; i<NUM_OUTS; i++)
		mOut[i]->save(file);

	fputc(mKeyActive, file);
	fputc(mCtrlActive, file);

	for (i=0; i<NUM_CTRLS; i++)
		mCtrlMap[i]->save(file);

	for (i=0; i<NUM_KEYS; i++)
		mKeyMap[i]->save(file);

	fclose(file);

	mFilename = filename;

	mDirty = false;
}



// give us access to the UI listboxes
void Doc::init_controls(CListBox *key_ctrl, CButton *key_active, CListBox *ctrl_ctrl, CButton *ctrl_active)
{
	mKeyMapCtrl = key_ctrl;
	mKeyActiveCtrl = key_active;

	mCtrlMapCtrl = ctrl_ctrl;
	mCtrlActiveCtrl = ctrl_active;

	update_key_controls();
	update_ctrl_controls();
}



// updates the list of key mappings
void Doc::update_key_controls()
{
	mKeyMapCtrl->SetRedraw(false);
	
	int top = mKeyMapCtrl->GetTopIndex();
	int sel = mKeyMapCtrl->GetCurSel();

	mKeyMapCtrl->ResetContent();

	for (int i=0; i<NUM_KEYS; i++)
		mKeyMapCtrl->AddString(mKeyMap[i]->get_desc().c_str());

	mKeyMapCtrl->SetTopIndex(top);
	mKeyMapCtrl->SetCurSel(sel);

	mKeyMapCtrl->SetRedraw(true);

	mKeyActiveCtrl->SetCheck(mKeyActive);
}



// updates the list of controller mappings
void Doc::update_ctrl_controls()
{
	mCtrlMapCtrl->SetRedraw(false);
	
	int top = mCtrlMapCtrl->GetTopIndex();
	int sel = mCtrlMapCtrl->GetCurSel();

	mCtrlMapCtrl->ResetContent();

	for (int i=0; i<NUM_CTRLS; i++)
		mCtrlMapCtrl->AddString(mCtrlMap[i]->get_desc().c_str());

	mCtrlMapCtrl->SetTopIndex(top);
	mCtrlMapCtrl->SetCurSel(sel);

	mCtrlMapCtrl->SetRedraw(true);

	mCtrlActiveCtrl->SetCheck(mCtrlActive);
}



// implements the zone crossfade
static int crossfade(int v, int p1, int p2, int p3, int p4, int fade)
{
	int prev = (p1 > 0)   ? (p1 + p2) / 2 : p1;
	int low  = (p2 > 0)   ? (p2 + p3) / 2 : p2;
	int high = (p3 < 256) ? (p2 + p3) / 2 : p3;
	int next = (p4 < 256) ? (p3 + p4) / 2 : p4;

	p1 = (prev*fade + p2*(256-fade)) / 256;
	p2 = (low *fade + p2*(256-fade)) / 256;
	p3 = (high*fade + p3*(256-fade)) / 256;
	p4 = (next*fade + p3*(256-fade)) / 256;

	if ((v >= p2) && (v < p3))
		return 256;

	if ((v < p1) || (v >= p4))
		return 0;

	if ((v >= p1) && (v < p2))
		return (v - p1) * 256 / (p2 - p1);

	if ((v >= p3) && (v < p4))
		return (p4 - v) * 256 / (p4 - p3);

	ASSERT(FALSE);

	return 0;
}



// process a MIDI note on message
void Doc::process_note_on(int channel, int key, int vel, int port)
{
	// stop any previous notes, and get a copy of the remembrance table
	process_note_off(channel, key, 0, port);

	NotesOn *n = &mNotesOn[port][key];

	ASSERT(n->mN == 0);

	if ((port == 0) && (mCtrlActive)) {
		// send pitch and vel through the controller mappings
		mLastNote = key;

		process_controller(channel, CtrlMap::KEY, key);
		process_controller(channel, CtrlMap::VELOCITY, vel);

		key = mCtrlOut[CtrlMap::KEY];
		vel = mCtrlOut[CtrlMap::VELOCITY];
	}

	if ((port > 0) || (!mKeyActive)) {
		// non-translated mode
		int out_port = (port > 0) ? MAX(mThruPort, 0) : 0;

		n->add(out_port, key);

		mOut[out_port]->out(MIDI_NOTE_ON | channel | (key << 8) | (vel << 16));
	}
	else {
		// look up according to the note mapping
		KeyMap *map = mKeyMap[key];

		// read controller values
		int c[2];

		for (int i=0; i<2; i++) {
			c[i] = mCtrlIn[map->mSplitController[i]];

			int min, max;

			CtrlMap::get_range(map->mSplitController[i], min, max);

			c[i] -= min;
			c[i] *= 255;

			if (min != max)
				c[i] /= (max - min);
		}

CString s;
s.Format("\ncontroller %d\n", c[0]);
OutputDebugString(s);

		// check each zone along the first axis
		for (int x=0; x<KeyMap::MAX_SPLITS; x++) {
			if (map->mSplitPoint[0][x] >= 256)
				break;

			int v1 = crossfade(c[0],
			                   map->mSplitPoint[0][MAX(x-1, 0)],
							   map->mSplitPoint[0][x],
							   map->mSplitPoint[0][x+1],
							   map->mSplitPoint[0][MIN(x+2, KeyMap::MAX_SPLITS)],
							   map->mSplitCrossfade[0]);

			if (v1 <= 0)
				continue;

			// check each zone along the second axis
			for (int y=0; y<KeyMap::MAX_SPLITS; y++) {
				if (map->mSplitPoint[1][y] >= 256)
					break;

				int v2 = crossfade(c[1],
								   map->mSplitPoint[1][MAX(y-1, 0)],
								   map->mSplitPoint[1][y],
								   map->mSplitPoint[1][y+1],
								   map->mSplitPoint[1][MIN(y+2, KeyMap::MAX_SPLITS)],
								   map->mSplitCrossfade[1]);

				if (v2 <= 0)
					continue;

				// look up the entry
				KeyMap::Entry *entry = &map->mMap[x][y];

				if ((entry->mPort >= 0) && (entry->mKey >= 0)) {

					// adjust velocity
					int v = (vel * entry->mVelScale / 256) + entry->mVelOffset;

					v = (v*v1) / 256;
					v = (v*v2) / 256;

					if (v < 0)
						v = 0;
					else if (v > 127)
						v = 127;

					// phew, play the note!
					if (v > 0) {
						n->add(entry->mPort, entry->mKey);

						mOut[entry->mPort]->out(MIDI_NOTE_ON | channel | (entry->mKey << 8) | (v << 16));
					}
				}
			}
		}
	}
}



// process a MIDI note off message
void Doc::process_note_off(int channel, int key, int vel, int port)
{
	NotesOn *n = &mNotesOn[port][key];

	for (int i=0; i<n->mN; i++)
		mOut[n->mPort[i]]->out(MIDI_NOTE_OFF | channel | (n->mKey[i] << 8) | (vel << 16));

	n->mN = 0;
}



// turns a pitch bend amount into MIDI format
static int encode_pitch_bend(int value)
{
	value += 0x2000;

	int d1 = value & 0x7F;
	int d2 = (value >> 7) & 0x7F;
	
	return (d1 << 8) | (d2 << 16);
}



// process a MIDI controller message
void Doc::process_controller(int channel, int ctrl, int value)
{
	mCtrlIn[ctrl] = value;
	
	CtrlMap *map = mCtrlMap[ctrl];

	for (int i=0; i<CtrlMap::NUM_MAPS; i++) {
		int c = map->mOut[i];

		if (c < 0)
			continue;

		// adjust range
		int v = value;

		if (map->mInMin[i] < map->mInMax[i]) {
			if (v < map->mInMin[i])
				v = map->mInMin[i];
			else if (v > map->mInMax[i])
				v = map->mInMax[i];
		}
		else {
			if (v > map->mInMin[i])
				v = map->mInMin[i];
			else if (v < map->mInMax[i])
				v = map->mInMax[i];
		}

		v -= map->mInMin[i];

		v *= (map->mOutMax[i] - map->mOutMin[i]);

		if (map->mInMin[i] != map->mInMax[i])
			v /= (map->mInMax[i] - map->mInMin[i]);

		v += map->mOutMin[i];

		int min, max;

		map->get_range(map->mOut[i], min, max);

		if (v < min)
			v = min;
		else if (v > max)
			v = max;

		mCtrlOut[c] = v;
		
		// output
		switch (c) {

			case CtrlMap::NONE:
				ASSERT(FALSE);
				break;

			case CtrlMap::KEY:
			case CtrlMap::VELOCITY:
				// don't do anything here: the note-on reads from mCtrlOut
				break;

			case CtrlMap::KEY_PRESSURE:
				mOut[0]->out(MIDI_KEY_PRESSURE | channel | (mLastNote << 8) | (v << 16));
				break;

			case CtrlMap::AFTERTOUCH:
				mOut[0]->out(MIDI_AFTERTOUCH | channel | (v << 8));
				break;

			case CtrlMap::PITCH_BEND:
				mOut[0]->out(MIDI_PITCH_BEND | channel | encode_pitch_bend(v));
				break;

			default:
				mOut[0]->out(MIDI_CONTROLLER | channel | ((c - CtrlMap::CONTROLLER) << 8) | (v << 16));
				break;
		}
	}
}



// processes MIDI input
void Doc::process(int msg, int channel, int d1, int d2, int port)
{
	if (msg == MIDI_NOTE_ON) {
		// note on
		process_note_on(channel, d1, d2, port);
	}
	else if (msg == MIDI_NOTE_OFF) {
		// note off
		process_note_off(channel, d1, d2, port);
	}
	else if ((port > 0) || (!mCtrlActive)) {
		// no filtering
		mOut[0]->out(msg | channel | (d1 << 8) | (d2 << 16));
	}
	else {
		// control mapping
		switch (msg) {

			case MIDI_KEY_PRESSURE:
				process_controller(channel, CtrlMap::KEY_PRESSURE, d2);
				break;

			case MIDI_CONTROLLER:
				process_controller(channel, CtrlMap::CONTROLLER + d1, d2);
				break;

			case MIDI_AFTERTOUCH:
				process_controller(channel, CtrlMap::AFTERTOUCH, d1);
				break;

			case MIDI_PITCH_BEND:
				process_controller(channel, CtrlMap::PITCH_BEND, ((d2 << 7) | d1) - 0x2000);
				break;
		}
	}
}


